package jp.takayuki.rtf;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Method;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import jp.takayuki.tools.ByteList;
import jp.takayuki.tools.Tool;

/**
 * RTF parser
 * <p>
 * plain textpicturê݂𒊏o܂B
 * <p>
 * s\ngp܂B
 * 
 * @author Takayuki Shoji
 */
public class RTFParser {

	/** Control symbol */
	private static final String SYMBOLS = "'*-:\\_{|}~"; //$NON-NLS-1$

	/** Control symbol̔z */
	private static final String[] SYMBOLS_ARRAY;
	static {
		SYMBOLS_ARRAY = new String[SYMBOLS.length()];
		for (int i = 0; i < SYMBOLS.length(); i++) {
			SYMBOLS_ARRAY[i] = "" + SYMBOLS.charAt(i); //$NON-NLS-1$
		}
	}

	/** s */
	private final String EOL;

	/** EOF(-1) */
	private static final int EOF = -1;

	/** {@link #ungetc(int)}̃obt@ */
	private int buffer = -2;

	/** plain text */
	private StringBuffer text = new StringBuffer();

	/** eLXgp */
	private ByteList byteList = new ByteList();

	/** ̓Xg[ */
	private InputStream in;

	/** astertrueɂȂ */
	private boolean ignorable;

	/** O[ṽX^bN */
	private LinkedList stack = new LinkedList();

	/**
	 * Control word/symbolɑ΂ANV
	 */
	private static final HashMap METHOD_MAP = new HashMap();
	static {
		// Control symbol
		try {
			registerMethod("'", "doSymbol_quote"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod("*", "doSymbol_aster"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod("-", "doSymbol_minus"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod(":", "doSymbol_colon"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod("\\", "doSymbol_backslash"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod("_", "doSymbol_underbar"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod("{", "doSymbol_leftbrace"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod("|", "doSymbol_pipe"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod("}", "doSymbol_rightbrace"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod("~", "doSymbol_tilde"); //$NON-NLS-1$ //$NON-NLS-2$

			registerMethod("colortbl", "skipGroup"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod("fonttbl", "skipGroup"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod("stylesheet", "skipGroup"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod("info", "skipGroup"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod("footnote", "skipGroup"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod("header", "skipGroup"); //$NON-NLS-1$ //$NON-NLS-2$
			registerMethod("footer", "skipGroup"); //$NON-NLS-1$ //$NON-NLS-2$

		} catch (SecurityException e) {
			throw e;
		} catch (NoSuchMethodException e) {
			throw new RuntimeException(e);
		}
		// Control word
		Method[] methods = RTFParser.class.getDeclaredMethods();
		for (int i = 0; i < methods.length; i++) {
			if (methods[i].getName().startsWith("doWord_")) { //$NON-NLS-1$
				METHOD_MAP.put(methods[i].getName().substring(7), methods[i]);
			}
		}
	}

	/**
	 * Rg[V{̃\bho^܂B
	 * @param key Control word, Control symbol
	 * @param name \bh
	 * @throws NoSuchMethodException
	 */
	private static void registerMethod(String key, String name)
			throws NoSuchMethodException {
		Method method = RTFParser.class.getDeclaredMethod(name, null);
		METHOD_MAP.put(key, method);
	}

	/**
	 * RXgN^
	 * @param in ̓Xg[
	 * @param eol sR[h
	 */
	public RTFParser(InputStream in, String eol) {
		this.in = in;
		EOL = eol;
		RTFState state = new RTFState();
		stack.addFirst(state);
	}

	/**
	 * Control symbolɑΉ郁\bhĂяo܂B
	 * @param name Control symbol܂Control wordB\͊܂܂ȂB
	 * @throws IOException ͗O
	 */
	private void invokeAction(String name) throws IOException {
		Method method = (Method) METHOD_MAP.get(name);
		if (method == null) {
			//System.out.println("ignored word=" + name);
			if (ignorable) {
				// O[vǂݔ΂
				skipGroup();
				ignorable = false;
			} else {
				// p[^ǂݔ΂
				readParameter(true, 0);
				//peek().destination = DESTINATION_SKIP;
			}
			return;
		}
		try {
			//System.out.println("control word=" + name);
			method.invoke(this, null);
		} catch (Exception e) {
			throw Tool.newIOException("Invoke failed.", e); //$NON-NLS-1$
		}
	}

	/**
	 * ͂s܂B
	 * @param in ̓Xg[ 
	 * @throws IOException 
	 */
	public void parse() throws IOException {
		loop: while (true) {
			int c = getc();
			switch (c) {
			case EOF:
				break loop;
			case '{':
				//System.out.println("group: start");
				push();
				break;
			case '}':
				//System.out.println("group: end");
				pop();
				break;
			case '\\':
				String word = readControl();
				invokeAction(word);
				break;
			case 0x00:
			case 0x0A:
			case 0x0D:
				break;
			default:
				parseText(c);
				break;
			}
		}
		flushByteList();
	}

	/**
	 * ܂B
	 * @param c Oɓǂݍ1oCg
	 * @throws IOException ͗O
	 */
	private void parseText(int c) throws IOException {
		switch (peek().textState) {
		case TEXT_NORMAL: {
			if (peek().destination != DESTINATION_SKIP) {
				appendText((char) c);
			}
			break;
		}
		case TEXT_BIN: {
			addByteToDestination(c);
			break;
		}
		case TEXT_HEX: {
			ungetc(c);
			int b = readHex();
			if (b == EOF) {
				throw new IOException("Unexpected EOF."); //$NON-NLS-1$
			}
			addByteToDestination(b);
			break;
		}
		}
	}

	/**
	 * ݂DestinationɃf[^ǉ܂B
	 * 
	 * @param c f[^
	 */
	private void addByteToDestination(int c) {
		switch (peek().destination) {
		case DESTINATION_TEXT:
			byteList.add((byte) c);
			break;
		case DESTINATION_PICT:
			peekPicture().data.add((byte) c);
			break;
		case DESTINATION_SKIP:
		default:
			break;
		}
	}

	/**
	 * X^bN̐擪𕡐Ēǉ܂B
	 */
	private void push() {
		RTFState next = new RTFState(peek());
		stack.addFirst(next);
	}

	/**
	 * X^bN̐擪폜܂B
	 */
	private void pop() {
		//RTFState saved = (RTFState) 
		stack.removeFirst();
		// O[vI
	}

	/**
	 * ǉ܂B
	 * @param str ǉ镶
	 */
	private void appendText(String str) {
		if (!byteList.isEmpty()) {
			flushByteList();
		}
		//System.out.println("appendText:'" + str + "'");
		text.append(str);
	}

	/**
	 * ǉ܂B
	 * @param c ǉ镶
	 */
	private void appendText(char c) {
		if (!byteList.isEmpty()) {
			flushByteList();
		}
		text.append(c);
	}

	/**
	 * byteListtbV܂B
	 */
	private void flushByteList() {
		if (!byteList.isEmpty()) {
			String s = byteList.toString(peek().cs);
			byteList.clear();
			//System.out.println("flushByteList:" + s);
			text.append(s);
		}
	}

	/**
	 * Control word܂Control symbolǂݍ݂܂B
	 * @return Control word܂Control symbolBEOF̏ꍇ͋󕶎B
	 * @throws IOException ͗O
	 */
	private String readControl() throws IOException {

		int next = getc();
		int index = indexOfSymbol(next);
		if (index != -1) {
			return SYMBOLS_ARRAY[index];
		}
		ungetc(next);

		StringBuffer sb = new StringBuffer();
		while (true) {
			int c = getc();
			if (c == EOF) {
				return sb.toString();
			}
			if (!isControlWord(c)) {
				ungetc(c);
				return sb.toString();
			}
			sb.append((char) c);
		}
	}

	/**
	 * p[^ǂݍ݂܂B
	 * @param skipDelim f~^œǂݎ̂ĂꍇtrueBȕꍇfalsew肷
	 * @param defaultValue ftHgl
	 * @return p[^Bꍇ͋󕶎B
	 * @throws IOException ͗O
	 */
	private int readParameter(boolean skipDelim, int defaultValue)
			throws IOException {
		int c = getc();
		if (c != '-' && !isdigit(c)) {
			if (!skipDelim || c != ' ') {
				ungetc(c);
			}
			return defaultValue;
		}
		StringBuffer sb = new StringBuffer(11);
		sb.append((char) c);
		while (true) {
			c = getc();
			if (c == EOF) {
				break;
			} else if (isdigit(c)) {
				sb.append((char) c);
				continue;
			} else if (!skipDelim || c != ' ') {
				ungetc(c);
			}
			break;
		}
		try {
			return Integer.parseInt(sb.toString());
		} catch (NumberFormatException e) {
			return defaultValue;
		}
	}

	/**
	 * 2oCg16iƉ肵ēǂݍ݁AintԂ܂B
	 * @return byteւ̕ϊʂ܂EOF
	 * @throws IOException ͗O
	 */
	private int readHex() throws IOException {
		int c = getc();
		if (c == EOF) {
			return EOF;
		}
		char[] chars = new char[2];
		chars[0] = (char) c;

		c = getc();
		if (c == EOF) {
			return EOF;
		}
		chars[1] = (char) c;
		String hex = new String(chars);
		return Integer.parseInt(hex, 16);
	}

	/**
	 * O[vǂݔ΂܂B
	 * @throws IOException ͗O
	 */
	private void skipGroup() throws IOException {
		int open = 1;
		while (true) {
			int c = getc();
			switch (c) {
			case EOF: // EOF
				break;
			case '\\':
				getc();// ̂Ă
				break;
			case '{':
				open++;
				break;
			case '}':
				open--;
				if (open <= 0) {
					// I
					ungetc(c);
					return;
				}
				break;
			}
		}
	}

	/**
	 * 1ǂݍ݂܂B
	 * @return 1
	 * @throws IOException ̓G[
	 */
	private int getc() throws IOException {
		if (buffer == -2) {
			return in.read();
		}
		int buf = buffer;
		buffer = -2;
		return buf;
	}

	/**
	 * 1߂܂B
	 * @param c 1
	 * @throws IOException ͗O
	 */
	private void ungetc(int c) throws IOException {
		if (buffer != -2) {
			throw new IOException("Already ungotten. old=" + buffer //$NON-NLS-1$
					+ ", new=" + c); //$NON-NLS-1$
		}
		buffer = c;
	}

	/**
	 * Control word̕ǂԂ܂B
	 * @param c 1
	 * @return Control word̕Ȃtrue
	 */
	private static boolean isControlWord(int c) {
		return ('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z');
	}

	/**
	 * Control symbol̕ǂԂ܂B
	 * @param c 1
	 * @return Control symbol̕Ȃtrue
	 */
	private static int indexOfSymbol(int c) {
		return SYMBOLS.indexOf(c);
	}

	/**
	 * \' A hexadecimal value, based on the specified character set (may be used to identify 8-bit values).
	 * @throws IOException 
	 */
	void doSymbol_quote() throws IOException {
		int b = readHex();
		if (b != EOF && peek().textState != TEXT_IGNORE) {
			byteList.add((byte) b);
		}
	}

	/**
	 * \* Marks a destination whose text should be ignored if not understood by the RTF reader.
	 */
	void doSymbol_aster() {
		ignorable = true;
	}

	/**
	 * \- (Optional hyphen)->Soft hyphen
	 */
	void doSymbol_minus() {
		appendText('\u00AD');
	}

	/**
	 * \:
	 * <p>
	 * Specifies a subentry in an index entry.
	 * @throws IOException ͗O
	 */
	void doSymbol_colon() throws IOException {
		// O[v𖳎
		skipGroup();
	}

	/**
	 * \\
	 */
	void doSymbol_backslash() {
		appendText('\\');
	}

	/**
	 * \_ (Non-breaking hyphen)
	 */
	void doSymbol_underbar() {
		appendText('\u2011');
	}

	/**
	 * \{
	 */
	void doSymbol_leftbrace() {
		appendText('{');
	}

	/**
	 * \| Formula character
	 * @throws IOException ͗O
	 */
	void doSymbol_pipe() throws IOException {
		// O[v𖳎
		skipGroup();
	}

	/**
	 * \}
	 */
	void doSymbol_rightbrace() {
		appendText('}');
	}

	/**
	 * \~ (Non-breaking space)
	 */
	void doSymbol_tilde() {
		appendText('\u00A0');
	}

	/**
	 * rtf
	 * <p>
	 * RTF Version
	 * @throws IOException ͗O
	 */
	void doWord_rtf() throws IOException {
		int version = readParameter(true, 1);
		if (version != 1) {
			throw new IOException("Unsupported version."); //$NON-NLS-1$
		}
	}

	/**
	 * ansi
	 * <p>
	 * Character Set
	 * @throws IOException
	 */
	void doWord_ansicpg() throws IOException {
		flushByteList();

		int cpg = readParameter(true, 0);
		String csName;
		switch (cpg) {
		case 437: // United States IBM
			csName = "IBM437"; //$NON-NLS-1$
			break;
		case 708: // Arabic (ASMO 708)
		case 709: // Arabic (ASMO 449+, BCON V4)
		case 710: // Arabic (transparent Arabic)
		case 711: // Arabic (Nafitha Enhanced)
		case 720: // Arabic (transparent ASMO)
			// not supported
			csName = "windows-1256"; //$NON-NLS-1$
			break;
		case 819: // Windows 3.1:	// (United States and Western Europe)
			csName = "windows-1250"; //$NON-NLS-1$
			break;
		case 850: // IBM multilingual
			csName = "IBM00858"; //$NON-NLS-1$
			break;
		case 852: // Eastern European
			csName = "IBM852"; //$NON-NLS-1$
			break;
		case 860: // Portuguese
			csName = "IBM860"; //$NON-NLS-1$
			break;
		case 862: // Hebrew
			csName = "IBM862"; //$NON-NLS-1$
			break;
		case 863: // French Canadian
			csName = "IBM863"; //$NON-NLS-1$
			break;
		case 864: // Arabic
			csName = "IBM864"; //$NON-NLS-1$
			break;
		case 865: // Norwegian
			csName = "IBM865"; //$NON-NLS-1$
			break;
		case 866: // Soviet Union
			csName = "IBM866"; //$NON-NLS-1$
			break;
		case 874: // Thai
			csName = "x-windows-874"; //$NON-NLS-1$
			break;
		case 932: // Japanese
			csName = "windows-31j"; //$NON-NLS-1$
			break;
		case 936: // Simplified Chinese
			csName = "x-mswin-936"; //$NON-NLS-1$
			break;
		case 949: // Korean
			csName = "x-windows-949"; //$NON-NLS-1$
			break;
		case 950: // Traditional Chinese
			csName = "x-windows-950"; //$NON-NLS-1$
			break;
		case 1250: // Eastern European
			csName = "windows-1250"; //$NON-NLS-1$
			break;
		case 1251: // Cyrillic
			csName = "windows-1251"; //$NON-NLS-1$
			break;
		case 1252: // Western European
			csName = "windows-1252"; //$NON-NLS-1$
			break;
		case 1253: // Greek
			csName = "windows-1253"; //$NON-NLS-1$
			break;
		case 1254: // Turkish
			csName = "windows-1254"; //$NON-NLS-1$
			break;
		case 1255: // Hebrew
			csName = "windows-1255"; //$NON-NLS-1$
			break;
		case 1256: // Arabic
			csName = "windows-1256"; //$NON-NLS-1$
			break;
		case 1257: // Baltic
			csName = "windows-1257"; //$NON-NLS-1$
			break;
		case 1258: // Vietnamese
			csName = "windows-1258"; //$NON-NLS-1$
			break;
		case 1361: // Johab
			csName = "x-Johab"; //$NON-NLS-1$
			break;
		case 10000: // MAC Roman
			csName = "x-MacRoman"; //$NON-NLS-1$
			break;
		case 10001: // TODO MAC Japan
			csName = "windows-31j"; //$NON-NLS-1$
			break;
		case 10004: // MAC Arabic
			csName = "x-MacArabic"; //$NON-NLS-1$
			break;
		case 10005: // MAC Hebrew
			csName = "x-MacHebrew"; //$NON-NLS-1$
			break;
		case 10006: // MAC Greek
			csName = "x-MacGreek"; //$NON-NLS-1$
			break;
		case 10007: // MAC Cyrillic
			csName = "x-MacCyrillic"; //$NON-NLS-1$
			break;
		case 10029: // MAC Latin2
			csName = "x-MacCentralEurope"; //$NON-NLS-1$
			break;
		case 10081: // MAC Turkish
			csName = "x-MacTurkish"; //$NON-NLS-1$
			break;
		case 57002: // Devanagari
		case 57003: // Bengali
		case 57004: // Tamil
		case 57005: // Telugu
		case 57006: // Assamese
		case 57007: // Oriya
		case 57008: // Kannada
		case 57009: // Malayalam
		case 57010: // Gujarati
		case 57011: // Punjabi
			// not supported
		default:
			csName = System.getProperty("file.encoding"); //$NON-NLS-1$
			break;
		}
		try {
			peek().cs = Charset.forName(csName);
		} catch (Exception e) {
			peek().cs = Tool.defaultCharset();
		}
	}

	/**
	 * ucN
	 * <p>
	 * uN ̖̎oCg
	 * @throws IOException ͗O
	 */
	void doWord_uc() throws IOException {
		peek().uc = readParameter(true, 1);
	}

	/**
	 * peek
	 * @return 擪̏
	 */
	private RTFState peek() {
		return (RTFState) stack.getFirst();
	}

	/**
	 * u
	 * @throws IOException ͗O
	 */
	void doWord_u() throws IOException {
		char code = (char) readParameter(true, 0x20);
		appendText(code);

		int uc = peek().uc;
		int saved = peek().textState;
		peek().textState = TEXT_IGNORE;
		for (int i = 0; i < uc; i++) {
			int c = getc();
			switch (c) {
			case '\\':
				// \'NN 1ƂĈ
				String word = readControl();
				invokeAction(word);
				break;
			default:
				//ignore
				break;
			}

		}
		peek().textState = saved;
	}

	/**
	 * cell
	 * @throws IOException ͗O
	 */
	void doWord_cell() throws IOException {
		appendText('\t');
		readParameter(true, 0);
	}

	/**
	 * row
	 * @throws IOException ͗O
	 */
	void doWord_row() throws IOException {
		appendText(EOL);
		readParameter(true, 0);
	}

	/**
	 * par
	 * @throws IOException ͗O
	 */
	void doWord_par() throws IOException {
		appendText(EOL);
		readParameter(true, 0);
	}

	/**
	 * bliptag
	 * @throws IOException ͗O
	 */
	void doWord_bliptag() throws IOException {
		if (peek().shppict == 0) {
			return;
		}
		RTFPicture pict = peekPicture();
		if (pict != null) {
			int tag = readParameter(true, Integer.MIN_VALUE);
			if (pict.bliptag == tag) {
				skipGroup();
			} else {
				pict.bliptag = tag;
			}
		}
	}

	/**
	 * shppict
	 * @throws IOException ͗O
	 */
	void doWord_shppict() {
		ignorable = false;
		peek().shppict = shppict;
		addPicture(shppict);
	}

	/**
	 * nonshppict
	 */
	void doWord_nonshppict() {
		peek().shppict = nonshppict;
		addPicture(nonshppict);
	}

	/**
	 * pict
	 */
	void doWord_pict() {
		peek().textState = TEXT_HEX;
		peek().destination = DESTINATION_PICT;
		if (peek().shppict == 0) {
			addPicture(0);
		}
	}

	/**
	 * 摜Xgɒǉ܂B
	 * @param shppictType 0, shppict, nonshppict
	 */
	private void addPicture(int shppictType) {
		flushByteList();
		RTFPicture pict = new RTFPicture();
		pict.position = text.length();
		if (shppictType != nonshppict) {
			text.append('\uFFFC'); // OBJ
		}
		pict.shppict = shppictType;
		pictureList.add(pict);
	}

	/**
	 * emfblip
	 */
	void doWord_emfblip() {
		setPictType(PICT_TYPE_EMF);
	}

	/**
	 * pngblip
	 */
	void doWord_pngblip() {
		setPictType(PICT_TYPE_PNG);
	}

	/**
	 * jpegblip
	 */
	void doWord_jpegblip() {
		setPictType(PICT_TYPE_JPEG);
	}

	/**
	 * macpict
	 */
	void doWord_macpict() {
		setPictType(PICT_TYPE_MAC);
	}

	/**
	 * wmetafile
	 * @throws IOException ͗O
	 */
	void doWord_wmetafile() throws IOException {
		int type = readParameter(true, 0);
		setPictType(PICT_TYPE_WMF + type);
	}

	/**
	 * 摜^Cvݒ肵܂B
	 * @param pictType 摜^Cv
	 */
	private void setPictType(int pictType) {
		RTFPicture pict = peekPicture();
		if (pict != null) {
			pict.pictType = pictType;
		}
	}

	/** pict:摜Xg */
	ArrayList pictureList = new ArrayList();

	/**
	 * @return Ỏ摜
	 */
	private RTFPicture peekPicture() {
		if (pictureList.isEmpty()) {
			return null;
		}
		return (RTFPicture) pictureList.get(pictureList.size() - 1);
	}

	/**
	 * pict:Picture
	 */
	public static class RTFPicture {

		/** ߍ݈ʒu */
		public int position;

		/** Shape pict type: ʏ=0, shppict=1, nonshppict=2 */
		public int shppict;

		/** 摜̃^Cv(PICT_TYPE_*) */
		public int pictType;

		/** bliptag */
		public int bliptag;

		/** f[^ */
		public ByteList data = new ByteList();
	}

	/** pict:enhanced meta file */
	public static final int PICT_TYPE_EMF = 1;

	/** pict:PNG */
	public static final int PICT_TYPE_PNG = 2;

	/** pict:JPEG */
	public static final int PICT_TYPE_JPEG = 3;

	/** pict:Mac PICT */
	public static final int PICT_TYPE_MAC = 4;

	/**
	 * pict:Windows Metafile
	 * <p>
	 * p[^(1-8)Z(6-13)B
	 */
	public static final int PICT_TYPE_WMF = 5;

	/** RIS - Rtf Internal State: 0 */
	private static final int TEXT_NORMAL = 0;

	/** RIS - Rtf Internal State: 1 */
	private static final int TEXT_BIN = 1;

	/** RIS - Rtf Internal State: 2 */
	private static final int TEXT_HEX = 2;

	/** RIS - Rtf Internal State: 3 */
	private static final int TEXT_IGNORE = 3;

	/** RDS - Rtf Destination State: 0 */
	private static final int DESTINATION_TEXT = 0;

	/** RDS - Rtf Destination State: 1 */
	private static final int DESTINATION_SKIP = 1;

	/** RDS - Rtf Destination State: 2 */
	private static final int DESTINATION_PICT = 2;

	/**
	 * 
	 */
	static class RTFState {
		/** RXgN^ */
		public RTFState() {
			//
		}

		/**
		 * p
		 * @param state 
		 */
		public RTFState(RTFState state) {
			//this.chp = (CHP) save.chp.clone();
			uc = state.uc;
			ansicpg = state.ansicpg;
			cs = state.cs;

			destination = state.destination;
			textState = state.textState;
			shppict = state.shppict;
		}

		/** vpeB: Unicode֕ */
		int uc = 1;

		/** vpeB: R[hy[W */
		int ansicpg;

		/** vpeB: Zbg */
		Charset cs = Tool.defaultCharset();

		/** destination:  */
		int destination = DESTINATION_TEXT;

		/** GeBeB̓ǂݕ(bin, hex, normal) */
		int textState = TEXT_NORMAL;

		/** Shape pict type: ʏ=0, shppict=1, nonshppict=2 */
		int shppict;
	}

	/** Shape pict type: shppict=1 */
	private static final int shppict = 1;

	/** Shape pict type: nonshppict=2 */
	private static final int nonshppict = 2;

	/**
	 * ǂԂ܂B
	 * @param c 
	 * @return ̏ꍇtrue
	 */
	private static boolean isdigit(int c) {
		return '0' <= c && c <= '9';
	}

	/**
	 * @return {
	 */
	public String getText() {
		return text.toString();
	}

	/**
	 * @return RTFPicturẽXg
	 */
	public List getPictureList() {
		return pictureList;
	}
}
